# This file is part of the MapProxy project.
# Copyright (C) 2024 terrestris <https://terrestris.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import threading
import time

from io import BytesIO

from mapproxy.cache.mbtiles import MBTilesCache, MBTilesLevelCache
from mapproxy.cache.tile import Tile
from mapproxy.image import ImageSource
from mapproxy.test.helper import assert_files_in_dir, assert_permissions
from mapproxy.test.image import create_tmp_image_buf
from mapproxy.test.unit.test_cache_tile import TileCacheTestBase
from mapproxy.util.sqlite3 import sqlite3

tile_image = create_tmp_image_buf((256, 256), color='blue')
tile_image2 = create_tmp_image_buf((256, 256), color='red')


class TestMBTileCache(TileCacheTestBase):
    def setup_method(self):
        TileCacheTestBase.setup_method(self)
        self.cache = MBTilesCache(os.path.join(self.cache_dir, 'tmp.mbtiles'))

    def teardown_method(self):
        if self.cache:
            self.cache.cleanup()
        TileCacheTestBase.teardown_method(self)

    def test_default_coverage(self):
        assert self.cache.coverage is None

    def test_load_empty_tileset(self):
        assert self.cache.load_tiles([Tile(None)]) is True
        assert self.cache.load_tiles([Tile(None), Tile(None), Tile(None)]) is True

    def test_load_more_than_2000_tiles(self):
        # prepare data
        for i in range(0, 2010):
            assert self.cache.store_tile(Tile((i, 0, 10),  ImageSource(BytesIO(b'foo'))))

        tiles = [Tile((i, 0, 10)) for i in range(0, 2010)]
        assert self.cache.load_tiles(tiles)

    def test_timeouts(self):
        self.cache._db_conn_cache.db = sqlite3.connect(self.cache.mbtile_file, timeout=0.05)

        def block():
            # block database by delaying the commit
            db = sqlite3.connect(self.cache.mbtile_file)
            cur = db.cursor()
            stmt = "INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data) VALUES (?,?,?,?)"
            cur.execute(stmt, (3, 1, 1, '1234'))
            time.sleep(0.2)
            db.commit()

        try:
            assert self.cache.store_tile(self.create_tile((0, 0, 1))) is True

            t = threading.Thread(target=block)
            t.start()
            time.sleep(0.05)
            assert self.cache.store_tile(self.create_tile((0, 0, 1))) is False
        finally:
            t.join()

        assert self.cache.store_tile(self.create_tile((0, 0, 1))) is True


class TestMBTileCachePermissions(TileCacheTestBase):
    def setup_method(self):
        TileCacheTestBase.setup_method(self)
        self.cache = MBTilesCache(os.path.join(self.cache_dir, 'tmp.mbtiles'), file_permissions='700')

    def teardown_method(self):
        if self.cache:
            self.cache.cleanup()
        TileCacheTestBase.teardown_method(self)

    def test_permissions(self):
        assert_permissions(self.cache.mbtile_file, '700')


class TestMBTileLevelCache(TileCacheTestBase):
    always_loads_metadata = True

    def setup_method(self):
        TileCacheTestBase.setup_method(self)
        self.cache = MBTilesLevelCache(self.cache_dir)

    def test_default_coverage(self):
        assert self.cache.coverage is None

    def test_level_files(self):
        assert_files_in_dir(self.cache_dir, [])

        self.cache.store_tile(self.create_tile((0, 0, 1)))
        assert_files_in_dir(self.cache_dir, ['1.mbtiles'], glob='*.mbtiles')

        self.cache.store_tile(self.create_tile((0, 0, 5)))
        assert_files_in_dir(self.cache_dir, ['1.mbtiles', '5.mbtiles'], glob='*.mbtiles')

    def test_remove_level_files(self):
        self.cache.store_tile(self.create_tile((0, 0, 1)))
        self.cache.store_tile(self.create_tile((0, 0, 2)))
        assert_files_in_dir(self.cache_dir, ['1.mbtiles', '2.mbtiles'], glob='*.mbtiles')

        self.cache.remove_level_tiles_before(1, remove_all=True)
        assert_files_in_dir(self.cache_dir, ['2.mbtiles'], glob='*.mbtiles')

    def test_remove_level_tiles_before(self):
        self.cache.store_tile(self.create_tile((0, 0, 1)))
        self.cache.store_tile(self.create_tile((0, 0, 2)))

        assert_files_in_dir(self.cache_dir, ['1.mbtiles', '2.mbtiles'], glob='*.mbtiles')
        assert self.cache.is_cached(Tile((0, 0, 1)))

        self.cache.remove_level_tiles_before(1, timestamp=time.time() - 60)
        assert self.cache.is_cached(Tile((0, 0, 1)))

        self.cache.remove_level_tiles_before(1, timestamp=time.time() + 60)
        assert not self.cache.is_cached(Tile((0, 0, 1)))

        assert_files_in_dir(self.cache_dir, ['1.mbtiles', '2.mbtiles'], glob='*.mbtiles')
        assert self.cache.is_cached(Tile((0, 0, 2)))

    def test_bulk_store_tiles_with_different_levels(self):
        self.cache.store_tiles([
            self.create_tile((0, 0, 1)),
            self.create_tile((0, 0, 2)),
            self.create_tile((1, 0, 2)),
            self.create_tile((1, 0, 1)),
        ], dimensions=None)

        assert_files_in_dir(self.cache_dir, ['1.mbtiles', '2.mbtiles'], glob='*.mbtiles')
        assert self.cache.is_cached(Tile((0, 0, 1)))
        assert self.cache.is_cached(Tile((1, 0, 1)))
        assert self.cache.is_cached(Tile((0, 0, 2)))
        assert self.cache.is_cached(Tile((1, 0, 2)))



# This file is part of the MapProxy project.
# Copyright (C) 2010-2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import os
import time
from collections import defaultdict

try:
    import cPickle as pickle
except ImportError:
    import pickle

import pytest

from mapproxy.seed.seeder import TileWalker, SeedTask, SeedProgress
from mapproxy.cache.dummy import DummyLocker
from mapproxy.cache.tile import TileManager
from mapproxy.source.tile import TiledSource
from mapproxy.grid.tile_grid import tile_grid_for_epsg, TileGrid
from mapproxy.srs import SRS
from mapproxy.util.coverage import BBOXCoverage, GeomCoverage
from mapproxy.seed.config import before_timestamp_from_options, SeedConfigurationError
from mapproxy.seed.config import (
    LevelsList,
    LevelsRange,
    LevelsResolutionList,
    LevelsResolutionRange,
)
from mapproxy.seed.util import ProgressStore
from mapproxy.test.helper import TempFile


try:
    from shapely.wkt import loads as load_wkt

    load_wkt  # prevent lint warning
except ImportError:
    load_wkt = None


class MockSeedPool(object):

    def __init__(self):
        self.seeded_tiles = defaultdict(set)

    def process(self, tiles, progess):
        for x, y, level in tiles:
            self.seeded_tiles[level].add((x, y))


class MockCache(object):

    def is_cached(self, tile, dimensions=None):
        return False


class TestSeeder(object):

    def setup_method(self):
        self.grid = TileGrid(SRS(4326), bbox=[-180, -90, 180, 90])
        self.source = TiledSource(self.grid, None)
        self.tile_mgr = TileManager(
            self.grid, MockCache(), [self.source], "png", locker=DummyLocker()
        )
        self.seed_pool = MockSeedPool()

    def make_bbox_task(self, bbox, srs, levels):
        md = dict(name="", cache_name="", grid_name="")
        coverage = BBOXCoverage(bbox, srs)
        return SeedTask(
            md, self.tile_mgr, levels, refresh_timestamp=None, refresh_all=False, coverage=coverage
        )

    def make_geom_task(self, geom, srs, levels):
        md = dict(name="", cache_name="", grid_name="")
        coverage = GeomCoverage(geom, srs)
        return SeedTask(
            md, self.tile_mgr, levels, refresh_timestamp=None, refresh_all=False, coverage=coverage
        )

    def test_seed_full_bbox(self):
        task = self.make_bbox_task([-180, -90, 180, 90], SRS(4326), [0, 1, 2])
        seeder = TileWalker(task, self.seed_pool, handle_uncached=True)
        seeder.walk()

        assert len(self.seed_pool.seeded_tiles) == 3
        assert self.seed_pool.seeded_tiles[0] == set([(0, 0)])
        assert self.seed_pool.seeded_tiles[1] == set([(0, 0), (1, 0)])
        assert self.seed_pool.seeded_tiles[2] == set(
            [(0, 0), (1, 0), (2, 0), (3, 0), (0, 1), (1, 1), (2, 1), (3, 1)]
        )

    def test_seed_small_bbox(self):
        task = self.make_bbox_task([-45, 0, 180, 90], SRS(4326), [0, 1, 2])
        seeder = TileWalker(task, self.seed_pool, handle_uncached=True)
        seeder.walk()

        assert len(self.seed_pool.seeded_tiles) == 3
        assert self.seed_pool.seeded_tiles[0] == set([(0, 0)])
        assert self.seed_pool.seeded_tiles[1] == set([(0, 0), (1, 0)])
        assert self.seed_pool.seeded_tiles[2] == set([(1, 1), (2, 1), (3, 1)])

    def test_seed_small_bbox_iregular_levels(self):
        task = self.make_bbox_task([-45, 0, 180, 90], SRS(4326), [0, 2])
        seeder = TileWalker(task, self.seed_pool, handle_uncached=True)
        seeder.walk()

        assert len(self.seed_pool.seeded_tiles) == 2
        assert self.seed_pool.seeded_tiles[0] == set([(0, 0)])
        assert self.seed_pool.seeded_tiles[2] == set([(1, 1), (2, 1), (3, 1)])

    def test_seed_small_bbox_transformed(self):
        bbox = SRS(4326).transform_bbox_to(SRS(900913), [-45, 0, 179, 80])
        task = self.make_bbox_task(bbox, SRS(900913), [0, 1, 2])
        seeder = TileWalker(task, self.seed_pool, handle_uncached=True)
        seeder.walk()

        assert len(self.seed_pool.seeded_tiles) == 3
        assert self.seed_pool.seeded_tiles[0] == set([(0, 0)])
        assert self.seed_pool.seeded_tiles[1] == set([(0, 0), (1, 0)])
        assert self.seed_pool.seeded_tiles[2] == set([(1, 1), (2, 1), (3, 1)])

    @pytest.mark.skipif(not load_wkt, reason="Shapely not installed")
    def test_seed_with_geom(self):
        # box from 10 10 to 80 80 with small spike/corner to -10 60 (upper left)
        geom = load_wkt("POLYGON((10 10, 10 50, -10 60, 10 80, 80 80, 80 10, 10 10))")
        task = self.make_geom_task(geom, SRS(4326), [0, 1, 2, 3, 4])
        seeder = TileWalker(task, self.seed_pool, handle_uncached=True)
        seeder.walk()

        assert len(self.seed_pool.seeded_tiles) == 5
        assert self.seed_pool.seeded_tiles[0] == set([(0, 0)])
        assert self.seed_pool.seeded_tiles[1] == set([(0, 0), (1, 0)])
        assert self.seed_pool.seeded_tiles[2] == set([(1, 1), (2, 1)])
        assert self.seed_pool.seeded_tiles[3] == set(
            [(4, 2), (5, 2), (4, 3), (5, 3), (3, 3)]
        )
        assert len(self.seed_pool.seeded_tiles[4]) == 4 * 4 + 2

    @pytest.mark.skipif(not load_wkt, reason="Shapely not installed")
    def test_seed_with_res_list(self):
        # box from 10 10 to 80 80 with small spike/corner to -10 60 (upper left)
        geom = load_wkt("POLYGON((10 10, 10 50, -10 60, 10 80, 80 80, 80 10, 10 10))")

        self.grid = TileGrid(
            SRS(4326),
            bbox=[-180, -90, 180, 90],
            res=[360 / 256, 360 / 720, 360 / 2000, 360 / 5000, 360 / 8000],
        )
        self.tile_mgr = TileManager(
            self.grid, MockCache(), [self.source], "png", locker=DummyLocker()
        )
        task = self.make_geom_task(geom, SRS(4326), [0, 1, 2, 3, 4])
        seeder = TileWalker(task, self.seed_pool, handle_uncached=True)
        seeder.walk()

        assert len(self.seed_pool.seeded_tiles) == 5
        assert self.seed_pool.seeded_tiles[0] == set([(0, 0)])
        assert self.grid.grid_sizes[1] == (3, 2)
        assert self.seed_pool.seeded_tiles[1] == set([(1, 0), (1, 1), (2, 0), (2, 1)])
        assert self.grid.grid_sizes[2] == (8, 4)
        assert self.seed_pool.seeded_tiles[2] == set(
            [(4, 2), (5, 2), (4, 3), (5, 3), (3, 3)]
        )
        assert self.grid.grid_sizes[3] == (20, 10)
        assert len(self.seed_pool.seeded_tiles[3]) == 5 * 5 + 2

    def test_seed_full_bbox_continue(self):
        task = self.make_bbox_task([-180, -90, 180, 90], SRS(4326), [0, 1, 2])
        seed_progress = SeedProgress([(0, 1), (1, 2)])
        seeder = TileWalker(
            task, self.seed_pool, handle_uncached=True, seed_progress=seed_progress
        )
        seeder.walk()

        assert len(self.seed_pool.seeded_tiles) == 3
        assert self.seed_pool.seeded_tiles[0] == set([(0, 0)])
        assert self.seed_pool.seeded_tiles[1] == set([(0, 0), (1, 0)])
        assert self.seed_pool.seeded_tiles[2] == set([(2, 0), (3, 0), (2, 1), (3, 1)])


class TestLevels(object):

    def test_level_list(self):
        levels = LevelsList([-10, 3, 1, 3, 5, 7, 50])
        assert levels.for_grid(tile_grid_for_epsg(4326)) == [1, 3, 5, 7]

    def test_level_range(self):
        levels = LevelsRange([1, 5])
        assert levels.for_grid(tile_grid_for_epsg(4326)) == [1, 2, 3, 4, 5]

    def test_level_range_open_from(self):
        levels = LevelsRange([None, 5])
        assert levels.for_grid(tile_grid_for_epsg(4326)) == [0, 1, 2, 3, 4, 5]

    def test_level_range_open_to(self):
        levels = LevelsRange([13, None])
        assert levels.for_grid(tile_grid_for_epsg(4326)) == [13, 14, 15, 16, 17, 18, 19]

    def test_level_range_open_tos_range(self):
        levels = LevelsResolutionRange([1000, 100])
        assert levels.for_grid(tile_grid_for_epsg(900913)) == [8, 9, 10, 11]

    def test_res_range_open_from(self):
        levels = LevelsResolutionRange([None, 100])
        assert levels.for_grid(tile_grid_for_epsg(900913)) == [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
        ]

    def test_res_range_open_to(self):
        levels = LevelsResolutionRange([1000, None])
        assert levels.for_grid(tile_grid_for_epsg(900913)) == [
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
        ]

    def test_resolution_list(self):
        levels = LevelsResolutionList([1000, 100, 500])
        assert levels.for_grid(tile_grid_for_epsg(900913)) == [8, 9, 11]


class TestProgressStore(object):

    def test_load_empty(self):
        store = ProgressStore("doesnotexist.no_realy.txt")
        store.load()
        assert store.get(("foo", "bar", "baz")) is None

    def test_load_store(self):
        with TempFile(no_create=True) as tmp:
            with open(tmp, "wb") as f:
                f.write(pickle.dumps({("view", "cache", "grid"): [(0, 1), (2, 4)]}))
            store = ProgressStore(tmp)
            assert store.get(("view", "cache", "grid")) == [(0, 1), (2, 4)]
            assert store.get(("view", "cache", "grid2")) is None
            store.add(("view", "cache", "grid"), [])
            store.add(("view", "cache", "grid2"), [(0, 1)])
            store.write()

            store = ProgressStore(tmp)
            assert store.get(("view", "cache", "grid")) == []
            assert store.get(("view", "cache", "grid2")) == [(0, 1)]

    def test_load_broken(self):
        with TempFile(no_create=True) as tmp:
            with open(tmp, "wb") as f:
                f.write(b"##invaliddata")
                f.write(pickle.dumps({("view", "cache", "grid"): [(0, 1), (2, 4)]}))

            store = ProgressStore(tmp)
            assert store.status == {}


class TestRemoveBeforeTimetamp(object):

    def test_from_time(self):
        ts = before_timestamp_from_options({"time": "2010-12-01T20:12:00"})
        # we don't know the timezone this test will run
        assert (1291230720.0 - 14 * 3600) < ts < (1291230720.0 + 14 * 3600)

    def test_from_mtime(self):
        with TempFile() as tmp:
            os.utime(tmp, (12376512, 12376512))
            assert before_timestamp_from_options({"mtime": tmp}) == 12376512

    def test_from_mtime_missing_file(self):
        with pytest.raises(SeedConfigurationError):
            before_timestamp_from_options(
                {"mtime": "/tmp/does-not-exist-at-all,really"}
            )

    def test_from_empty(self):
        assert before_timestamp_from_options({}) == pytest.approx(time.time(), 1)

    def test_from_delta(self):
        assert (
            before_timestamp_from_options({"minutes": 15}) + 60 * 15
        ) == pytest.approx(time.time(), 1)


class TestSeedProgress(object):

    def test_progress_identifier(self):
        old = SeedProgress()
        with old.step_down(0, 2):
            with old.step_down(0, 4):
                assert old.current_progress_identifier() == [(0, 2), (0, 4)]
            # previous leafs are still present
            assert old.current_progress_identifier() == [(0, 2), (0, 4)]
            with old.step_down(1, 4):
                assert old.current_progress_identifier() == [(0, 2), (1, 4)]
            assert old.current_progress_identifier() == [(0, 2), (1, 4)]

        assert old.current_progress_identifier() == []  # empty list after seed

        with old.step_down(1, 2):
            assert old.current_progress_identifier() == [(1, 2)]
            with old.step_down(0, 4):
                with old.step_down(1, 4):
                    assert old.current_progress_identifier() == [(1, 2), (0, 4), (1, 4)]

    def test_already_processed(self):
        new = SeedProgress([(0, 2)])
        with new.step_down(0, 2):
            assert not new.already_processed()
            with new.step_down(0, 2):
                assert not new.already_processed()

        new = SeedProgress([(1, 2)])
        with new.step_down(0, 2):
            assert new.already_processed()
            with new.step_down(0, 2):
                assert new.already_processed()

        new = SeedProgress([(0, 2), (1, 4), (2, 4)])
        with new.step_down(0, 2):
            assert not new.already_processed()
            with new.step_down(0, 4):
                assert new.already_processed()
            with new.step_down(1, 4):
                assert not new.already_processed()
                with new.step_down(1, 4):
                    assert new.already_processed()
                with new.step_down(2, 4):
                    assert not new.already_processed()
                with new.step_down(3, 4):
                    assert not new.already_processed()
            with new.step_down(2, 4):
                assert not new.already_processed()



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from mapproxy.config import base_config
from mapproxy import srs, proj
from mapproxy.srs import SRS, PreferredSrcSRS, SupportedSRS


class TestSRS(object):

    def test_epsg4326(self):
        srs = SRS(4326)

        assert srs.is_latlong
        assert not srs.is_axis_order_en
        assert srs.is_axis_order_ne

        assert srs.get_geographic_srs() == srs

    def test_crs84(self):
        srs = SRS("CRS:84")

        assert srs.is_latlong
        assert srs.is_axis_order_en
        assert not srs.is_axis_order_ne

        assert srs == SRS("EPSG:4326")

    def test_epsg31467(self):
        srs = SRS("EPSG:31467")

        assert not srs.is_latlong
        assert not srs.is_axis_order_en
        assert srs.is_axis_order_ne

        assert srs.get_geographic_srs() == SRS(4326)

    def test_epsg900913(self):
        srs = SRS("epsg:900913")

        assert not srs.is_latlong
        assert srs.is_axis_order_en
        assert not srs.is_axis_order_ne

    def test_non_epsg_auth(self):
        srs = SRS("IGNF:ETRS89UTM28")

        assert not srs.is_latlong
        assert srs.is_axis_order_en
        assert not srs.is_axis_order_ne

        assert srs.get_geographic_srs() == SRS(4326)

    def test_iau_crs(self):
        try:
            srs = SRS("IAU:49900")  # "Mars (2015) - Sphere / Ocentric"
        except Exception:
            pytest.skip('Requires recent pyproj version')

        assert srs.is_latlong
        assert '49900' in str(srs.get_geographic_srs())

    def test_iau_crs_projected(self):
        try:
            srs = SRS("IAU:49910")  # "Mars (2015) - Sphere / Ocentric / Equirectangular, clon = 0"
        except Exception:
            pytest.skip('Requires recent pyproj version')

        assert not srs.is_latlong
        assert '49900' in str(srs.get_geographic_srs())

    def test_from_srs(self):
        srs1 = SRS("epsg:4326")
        srs2 = SRS(srs1)
        assert srs1 == srs2

# proj_data_dir test relies on old Proj4 epsg files.


@pytest.mark.skipif(not proj.USE_PROJ4_API, reason="only for old proj4 lib")
class Test_0_ProjDefaultDataPath(object):

    def test_known_srs(self):
        srs.SRS(4326)

    def test_unknown_srs(self):
        with pytest.raises(RuntimeError):
            srs.SRS(1234)


@pytest.fixture(scope="class")
def custom_proj_data_dir():
    srs._proj_initalized = False
    srs._srs_cache = {}
    base_config().srs.proj_data_dir = os.path.dirname(__file__)

    yield

    srs._proj_initalized = False
    srs._srs_cache = {}
    srs.set_datapath(None)
    base_config().srs.proj_data_dir = None


@pytest.mark.skipif(not proj.USE_PROJ4_API, reason="only for old proj4 lib")
@pytest.mark.usefixtures("custom_proj_data_dir")
class Test_1_ProjDataPath(object):

    def test_dummy_srs(self):
        srs.SRS(1234)

    def test_unknown_srs(self):
        with pytest.raises(RuntimeError):
            srs.SRS(2339)


class TestPreferredSrcSRS(object):

    # test selection of preferred SRS
    # unprojected: 4326, 4258
    # projected: 3857, 25831, 25832, 31467
    @pytest.mark.parametrize("target,available,expected", [

        # always return target if available
        (4326, [25832, 4326, 3857], 4326),

        (4326, [25832, 4326, 3857], 4326),
        (4326, [25831, 3857], 3857),

        (3857, [25832, 4258, 3857, 31466], 3857),
        (3857, [25832, 4258, 31467, 25831, 31466], 25832),
        (3857, [4258, 31467, 25831, 31466], 25831),
        (3857, [4258, 31467, 31466], 31467),
        (3857, [4258, 31466], 31466),
        (3857, [4258], 4258),

        # always return first preferred, regardless of order in available
        (4326, [3857, 4258], 4258),
        (4326, [4258, 3857], 4258),

        # no preferred, return first that is also projected/unprojected
        (31467, [4326, 25831, 3857], 25831),
        (4258, [25831, 4326, 3857], 4326),

        # no preferred and no srs that is also projected/unprojected, return first
        (31467, [4326, 4258], 4326),
        (4258, [3857, 25832, 31467], 3857),
    ])
    def test_preferred(self, target, available, expected):
        preferredSRS = PreferredSrcSRS()
        preferredSRS.add(SRS(4326), [SRS(4258), SRS(3857)])
        preferredSRS.add(SRS(3857), [SRS(25832), SRS(25831), SRS(31467), SRS(4326)])

        assert preferredSRS.preferred_src(SRS(target), [SRS(c) for c in available]) == SRS(expected)

    def test_no_available(self):
        preferredSRS = PreferredSrcSRS()
        preferredSRS.add(SRS(4326), [SRS(4258), SRS(3857)])

        with pytest.raises(ValueError):
            preferredSRS.preferred_src(SRS(4326), [])


class TestSupportedSRS(object):
    @pytest.fixture
    def preferred(self):
        preferredSRS = PreferredSrcSRS()
        preferredSRS.add(SRS(4326), [SRS(4258), SRS(3857)])
        preferredSRS.add(SRS(3857), [SRS(25832), SRS(25831), SRS(31467), SRS(4326)])
        return preferredSRS

    def test_supported(self, preferred):
        supported = SupportedSRS([SRS(4326), SRS(25832)], preferred)
        assert SRS(4326) in supported
        assert SRS(4258) not in supported
        assert SRS(25832) in supported

    def test_iter(self, preferred):
        supported = SupportedSRS([SRS(4326), SRS(25832)], preferred)
        assert [SRS(4326), SRS(25832)] == [srs for srs in supported]

    def test_best_srs(self, preferred):
        supported = SupportedSRS([SRS(4326), SRS(25832)], preferred)
        assert supported.best_srs(SRS(4326)) == SRS(4326)
        assert supported.best_srs(SRS(4258)) == SRS(4326)
        assert supported.best_srs(SRS(25832)) == SRS(25832)
        assert supported.best_srs(SRS(25831)) == SRS(25832)
        assert supported.best_srs(SRS(3857)) == SRS(25832)
        supported = SupportedSRS([SRS(4326), SRS(31467), SRS(25831)], preferred)
        assert supported.best_srs(SRS(3857)) == SRS(25831)
        assert supported.best_srs(SRS(25831)) == SRS(25831)

    def test_best_srs_no_preferred(self, preferred):
        supported = SupportedSRS([SRS(4326), SRS(25832)], None)
        assert supported.best_srs(SRS(4326)) == SRS(4326)
        assert supported.best_srs(SRS(4258)) == SRS(4326)
        assert supported.best_srs(SRS(25832)) == SRS(25832)
        assert supported.best_srs(SRS(25831)) == SRS(25832)
        assert supported.best_srs(SRS(3857)) == SRS(25832)


def test_ogc_crs_url_to_auth_code():

    assert srs.ogc_crs_url_to_auth_code("http://www.opengis.net/def/crs/EPSG/0/3857") == "EPSG:3857"

    with pytest.raises(ValueError):
        srs.ogc_crs_url_to_auth_code("invalid")



# SPDX-License-Identifier: MIT
# Copyright (C) pygeoapi developers

# Below is strongly derived from pygeoapi/util.py

import base64
import dateutil
import json
import pathlib
import uuid
from babel.support import Translations
from datetime import date, datetime, time
from decimal import Decimal
from importlib import resources as importlib_resources
from jinja2 import Environment, FileSystemLoader, select_autoescape
from jinja2.exceptions import TemplateNotFound
from pathlib import Path
from typing import Any

from mapproxy.config.config import base_config
from mapproxy.version import __version__

import logging

LOGGER = logging.getLogger(__name__)

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def json_serial(obj: Any) -> str:
    """
    helper function to convert to JSON non-default
    types (source: https://stackoverflow.com/a/22238613)

    :param obj: `object` to be evaluated

    :returns: JSON non-default type to `str`
    """

    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    elif isinstance(obj, bytes):
        try:
            LOGGER.debug("Returning as UTF-8 decoded bytes")
            return obj.decode("utf-8")
        except UnicodeDecodeError:
            LOGGER.debug("Returning as base64 encoded JSON object")
            return base64.b64encode(obj)
    elif isinstance(obj, Decimal):
        return float(obj)
    elif type(obj).__name__ in ["int32", "int64"]:
        return int(obj)
    elif type(obj).__name__ in ["float32", "float64"]:
        return float(obj)
    elif isinstance(obj, (pathlib.PurePath, Path)):
        return str(obj)
    elif isinstance(obj, uuid.UUID):
        return str(obj)
    else:
        msg = f"{obj} type {type(obj)} not serializable"
        LOGGER.error(msg)
        raise TypeError(msg)


def to_json(dict_: dict, pretty: bool = False) -> str:
    """
    Serialize dict to json

    :param dict_: `dict` of JSON representation
    :param pretty: `bool` of whether to prettify JSON (default is `False`)

    :returns: JSON string representation
    """

    if pretty:
        indent = 4
    else:
        indent = None

    return json.dumps(dict_, default=json_serial, indent=indent, separators=(",", ":"))


def format_datetime(value: str, format_: str = DATETIME_FORMAT) -> str:
    """
    Parse datetime as ISO 8601 string; re-present it in particular format
    for display in HTML

    :param value: `str` of ISO datetime
    :param format_: `str` of datetime format for strftime

    :returns: string
    """

    if not isinstance(value, str) or not value.strip():
        return ""

    return dateutil.parser.isoparse(value).strftime(format_)


def human_size(nbytes: int) -> str:
    """
    Provides human readable file size

    source: https://stackoverflow.com/a/14996816

    :param nbytes: int of file size (bytes)
    :param units: list of unit abbreviations

    :returns: string of human readable filesize
    """

    suffixes = ["B", "K", "M", "G", "T", "P"]

    i = 0

    while nbytes >= 1024 and i < len(suffixes) - 1:
        nbytes /= 1024.0
        i += 1

    if suffixes[i] == "K":
        f = str(int(nbytes)).rstrip("0").rstrip(".")
    elif suffixes[i] == "B":
        return nbytes
    else:
        f = f"{nbytes:.1f}".rstrip("0").rstrip(".")

    return f"{f}{suffixes[i]}"


def get_path_basename(urlpath: str) -> str:
    """
    Helper function to derive file basename

    :param urlpath: URL path

    :returns: string of basename of URL path
    """

    return Path(urlpath).name


def get_breadcrumbs(urlpath: str) -> list:
    """
    helper function to make breadcrumbs from a URL path

    :param urlpath: URL path

    :returns: `list` of `dict` objects of labels and links
    """

    links = []

    tokens = urlpath.split("/")

    s = ""
    for t in tokens:
        if s:
            s += "/" + t
        else:
            s = t
        links.append(
            {
                "href": s,
                "title": t,
            }
        )

    return links


def filter_dict_by_key_value(dict_: dict, key: str, value: str) -> dict:
    """
    helper function to filter a dict by a dict key

    :param dict_: ``dict``
    :param key: dict key
    :param value: dict key value

    :returns: filtered ``dict``
    """

    return {k: v for (k, v) in dict_.items() if v[key] == value}


def render_j2_template(
    config: dict,
    module_name: str,
    template_root: Path,
    template: Path,
    data: dict,
    locale_: str = None,
) -> str:
    """
    render Jinja2 template

    :param config: dict of configuration
    :param module_name: module name from which to get templates
    :param template_root: template (relative path to module_name)
    :param template: template (relative path to template_root)
    :param data: dict of data
    :param locale_: the requested output Locale

    :returns: string of rendered template
    """

    if base_config().template_dir:
        template_dir = base_config().template_dir + "/" + template_root
    else:
        template_dir = (
            importlib_resources.files(module_name)
            .joinpath("templates")
            .joinpath(template_root)
        )

    template_paths = [template_dir]

    env = Environment(
        loader=FileSystemLoader(template_paths),
        extensions=["jinja2.ext.i18n"],
        autoescape=select_autoescape(),
    )

    env.filters["to_json"] = to_json
    env.filters["format_datetime"] = format_datetime
    env.filters["human_size"] = human_size
    env.globals.update(to_json=to_json)

    env.filters["get_path_basename"] = get_path_basename
    env.globals.update(get_path_basename=get_path_basename)

    env.filters["get_breadcrumbs"] = get_breadcrumbs
    env.globals.update(get_breadcrumbs=get_breadcrumbs)

    env.filters["filter_dict_by_key_value"] = filter_dict_by_key_value
    env.globals.update(filter_dict_by_key_value=filter_dict_by_key_value)

    locale_dir = "."
    translations = Translations.load(locale_dir, [locale_])
    env.install_gettext_translations(translations)

    try:
        template = env.get_template(template)
    except TemplateNotFound:
        LOGGER.debug(f"template {template} not found")
        raise

    return template.render(
        config=config,
        data=data,
        locale=locale_,
        version=__version__,
    )



# -:- encoding: utf-8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from mapproxy.request import Request
from mapproxy.response import Response
from mapproxy.util.collections import LRU
from mapproxy.wsgiapp import make_wsgi_app as make_mapproxy_wsgi_app
from mapproxy.util.escape import escape_html

from threading import Lock

import logging
log = logging.getLogger(__name__)


def asbool(value):
    """
    >>> all([asbool(True), asbool('trUE'), asbool('ON'), asbool(1)])
    True
    >>> any([asbool(False), asbool('false'), asbool('foo'), asbool(None)])
    False
    """
    value = str(value).lower()
    return value in ('1', 'true', 'yes', 'on')


def app_factory(global_options, config_dir, allow_listing=False, **local_options):
    """
    Create a new MultiMapProxy app.

    :param config_dir: directory with all mapproxy configurations
    :param allow_listing: allow to list all available apps
    """
    return make_wsgi_app(config_dir, asbool(allow_listing))


def make_wsgi_app(config_dir, allow_listing=True, debug=False):
    """
    Create a MultiMapProxy with the given config directory.

    :param config_dir: the directory with all project configurations.
    :param allow_listing: True if MapProxy should list all instances
        at the root URL
    """
    config_dir = os.path.abspath(config_dir)
    loader = DirectoryConfLoader(config_dir)
    return MultiMapProxy(loader, list_apps=allow_listing, debug=debug)


class MultiMapProxy(object):

    def __init__(self, loader, list_apps=False, app_cache_size=100, debug=False):
        self.loader = loader
        self.list_apps = list_apps
        self._app_init_lock = Lock()
        self.apps = LRU(app_cache_size)
        self.debug = debug

    def __call__(self, environ, start_response):
        req = Request(environ)
        return self.handle(req)(environ, start_response)

    def handle(self, req):
        app_name = req.pop_path()
        if not app_name:
            return self.index_list(req)

        if not app_name or (
            app_name not in self.apps and not self.loader.app_available(app_name)
        ):
            return Response('not found', status=404)

        # safe instance/app name for authorization
        req.environ['mapproxy.instance_name'] = app_name
        return self.proj_app(app_name)

    def index_list(self, req):
        """
        Return greeting response with a list of available apps (if enabled with list_apps).
        """
        import mapproxy.version
        html = "<html><body><h1>Welcome to MapProxy %s</h1>" % mapproxy.version.version

        url = escape_html(req.script_url)
        if self.list_apps:
            html += "<h2>available instances:</h2><ul>"
            html += '\n'.join('<li><a href="%(url)s/%(name)s/">%(name)s</a></li>' % {'url': url, 'name': app}
                              for app in self.loader.available_apps())
            html += '</ul>'
        html += '</body></html>'
        return Response(html, content_type='text/html')

    def proj_app(self, proj_name):
        """
        Return the (cached) project app.
        """
        proj_app, timestamps = self.apps.get(proj_name, (None, None))

        if proj_app:
            if self.loader.needs_reload(proj_name, timestamps):
                # discard cached app
                proj_app = None

        if not proj_app:
            with self._app_init_lock:
                proj_app, timestamps = self.apps.get(proj_name, (None, None))
                if self.loader.needs_reload(proj_name, timestamps):
                    proj_app, timestamps = self.create_app(proj_name)
                    self.apps[proj_name] = proj_app, timestamps
                else:
                    proj_app, timestamps = self.apps[proj_name]

        return proj_app

    def create_app(self, proj_name):
        """
        Returns a new configured MapProxy app and a dict with the
        timestamps of all configuration files.
        """
        mapproxy_conf = self.loader.app_conf(proj_name)['mapproxy_conf']
        log.info('initializing project app %s with %s', proj_name, mapproxy_conf)
        app = make_mapproxy_wsgi_app(mapproxy_conf, debug=self.debug)
        return app, app.config_files


class ConfLoader(object):
    def needs_reload(self, app_name, timestamps):
        """
        Returns ``True`` if the configuration of `app_name` changed
        since `timestamp`.
        """
        raise NotImplementedError()

    def app_available(self, app_name):
        """
        Returns ``True`` if `app_name` is available.
        """
        raise NotImplementedError()

    def available_apps(self):
        """
        Returns a list with all available lists.
        """
        raise NotImplementedError()

    def app_conf(self, app_name):
        """
        Returns a configuration dict for the given `app_name`,
        None if the app is not found.

        The configuration dict contains at least 'mapproxy_conf'
        with the filename of the configuration.
        """
        raise NotImplementedError()


class DirectoryConfLoader(ConfLoader):
    """
    Load application configurations from a directory.
    """

    def __init__(self, base_dir, suffix='.yaml'):
        self.base_dir = base_dir
        self.suffix = suffix

    def needs_reload(self, app_name, timestamps):
        if not timestamps:
            return True
        for conf_file, timestamp in timestamps.items():
            m_time = os.path.getmtime(conf_file)
            if m_time > timestamp:
                return True
        return False

    def _is_conf_file(self, fname):
        if not os.path.isfile(fname):
            return False
        if self.suffix:
            return fname.lower().endswith(self.suffix)
        else:
            return True

    def app_name_from_filename(self, fname):
        """
        >>> DirectoryConfLoader('/tmp/').app_name_from_filename('/tmp/foobar.yaml')
        'foobar'
        """
        _path, fname = os.path.split(fname)
        app_name, _ext = os.path.splitext(fname)
        return app_name

    def filename_from_app_name(self, app_name):
        """
        >>> DirectoryConfLoader('/tmp/').filename_from_app_name('foobar')
        '/tmp/foobar.yaml'
        """
        return os.path.join(self.base_dir, app_name + self.suffix or '')

    def available_apps(self):
        apps = []
        for f in os.listdir(self.base_dir):
            if self._is_conf_file(os.path.join(self.base_dir, f)):
                app_name = self.app_name_from_filename(f)
                apps.append(app_name)
        apps.sort()
        return apps

    def app_available(self, app_name):
        conf_file = self.filename_from_app_name(app_name)
        return self._is_conf_file(conf_file)

    def app_conf(self, app_name):
        conf_file = self.filename_from_app_name(app_name)
        if not self._is_conf_file(conf_file):
            return None
        return {'mapproxy_conf': conf_file}
